@@ -13,16 +13,32 @@ import logging
 import sys
 import time
 import xml.etree.ElementTree as ET
+from collections import defaultdict
+from collections.abc import Hashable
 from io import BytesIO
-from typing import Awaitable, Callable, Dict, Hashable, Optional, Tuple, cast
+from typing import (
+    Awaitable,
+    Callable,
+    Coroutine,
+    DefaultDict,
+    Dict,
+    List,
+    Optional,
+    Set,
+    Tuple,
+    cast,
+)
 
 import attr
 import httpx
-from asyncstdlib import lru_cache
-from defusedxml.ElementTree import fromstring
+from defusedxml import DefusedXmlException
+from defusedxml.ElementTree import ParseError, fromstring
 
 from .appcommand import AppCommandCmd
 from .const import (
+    ALL_TELNET_EVENTS,
+    ALL_ZONE_TELNET_EVENTS,
+    ALL_ZONES,
     APPCOMMAND0300_URL,
     APPCOMMAND_CMD_TEXT,
     APPCOMMAND_NAME,
@@ -34,15 +50,12 @@ from .const import (
     ZONE2,
     ZONE3,
 )
-from .decorators import (
-    async_handle_receiver_exceptions,
-    cache_clear_on_exception,
-    set_cache_id,
-)
+from .decorators import async_handle_receiver_exceptions, cache_result
 from .exceptions import (
     AvrIncompleteResponseError,
     AvrInvalidResponseError,
     AvrNetworkError,
+    AvrProcessingError,
     AvrTimoutError,
 )
 
@@ -61,16 +74,96 @@ def get_default_async_client() -> httpx.AsyncClient:
     return httpx.AsyncClient()
 
 
-@attr.s(auto_attribs=True, hash=False, on_setattr=DENON_ATTR_SETATTR)
+def telnet_event_map_factory() -> Dict[str, List]:
+    """Create telnet event map."""
+    event_map: DefaultDict[str, List] = defaultdict(list)
+    for event in TELNET_EVENTS:
+        event_map[event[0:2]].append(event)
+    for value in event_map.values():
+        value.sort(key=len, reverse=True)
+    return dict(event_map)
+
+
+@attr.s(auto_attribs=True, hash=False)
+class HTTPXAsyncClient:
+    """Perform cached HTTP calls with httpx.AsyncClient."""
+
+    client_getter: Callable[[], httpx.AsyncClient] = attr.ib(
+        validator=attr.validators.is_callable(),
+        default=get_default_async_client,
+        init=False,
+    )
+
+    def __hash__(self) -> int:
+        """Hash the class using its ID that caching works."""
+        return id(self)
+
+    @cache_result
+    @async_handle_receiver_exceptions
+    async def async_get(
+        self,
+        url: str,
+        timeout: float,
+        read_timeout: float,
+        *,
+        cache_id: Hashable = None,
+    ) -> httpx.Response:
+        """Call GET endpoint of Denon AVR receiver asynchronously."""
+        client = self.client_getter()
+        try:
+            res = await client.get(
+                url, timeout=httpx.Timeout(timeout, read=read_timeout)
+            )
+            res.raise_for_status()
+        finally:
+            # Close the default AsyncClient but keep custom clients open
+            if self.is_default_async_client():
+                await client.aclose()
+
+        return res
+
+    @cache_result
+    @async_handle_receiver_exceptions
+    async def async_post(
+        self,
+        url: str,
+        timeout: float,
+        read_timeout: float,
+        *,
+        content: Optional[bytes] = None,
+        data: Optional[Dict] = None,
+        cache_id: Hashable = None,
+    ) -> httpx.Response:
+        """Call GET endpoint of Denon AVR receiver asynchronously."""
+        client = self.client_getter()
+        try:
+            res = await client.post(
+                url,
+                content=content,
+                data=data,
+                timeout=httpx.Timeout(timeout, read=read_timeout),
+            )
+            res.raise_for_status()
+        finally:
+            # Close the default AsyncClient but keep custom clients open
+            if self.is_default_async_client():
+                await client.aclose()
+
+        return res
+
+    def is_default_async_client(self) -> bool:
+        """Check if default httpx.AsyncClient getter is used."""
+        return self.client_getter is get_default_async_client
+
+
+@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
 class DenonAVRApi:
     """Perform API calls to Denon AVR REST interface."""
 
     host: str = attr.ib(converter=str, default="localhost")
     port: int = attr.ib(converter=int, default=80)
-    timeout: httpx.Timeout = attr.ib(
-        validator=attr.validators.instance_of(httpx.Timeout),
-        default=httpx.Timeout(2.0, read=15.0),
-    )
+    timeout: float = attr.ib(converter=float, default=2.0)
+    read_timeout: float = attr.ib(converter=float, default=15.0)
     _appcommand_update_tags: Tuple[AppCommandCmd] = attr.ib(
         validator=attr.validators.deep_iterable(
             attr.validators.instance_of(AppCommandCmd),
@@ -85,73 +178,53 @@ class DenonAVRApi:
         ),
         default=attr.Factory(tuple),
     )
-    async_client_getter: Callable[[], httpx.AsyncClient] = attr.ib(
-        validator=attr.validators.is_callable(),
-        default=get_default_async_client,
+    httpx_async_client: HTTPXAsyncClient = attr.ib(
+        validator=attr.validators.instance_of(HTTPXAsyncClient),
+        default=attr.Factory(HTTPXAsyncClient),
         init=False,
     )
 
-    def __hash__(self) -> int:
-        """
-        Hash the class in a custom way that caching works.
-
-        It should react on changes of host and port.
-        """
-        return hash((self.host, self.port))
-
-    @async_handle_receiver_exceptions
     async def async_get(
-        self, request: str, port: Optional[int] = None
+        self,
+        request: str,
+        *,
+        port: Optional[int] = None,
+        cache_id: Hashable = None,
     ) -> httpx.Response:
         """Call GET endpoint of Denon AVR receiver asynchronously."""
         # Use default port of the receiver if no different port is specified
         port = port if port is not None else self.port
 
-        endpoint = "http://{host}:{port}{request}".format(
-            host=self.host, port=port, request=request
+        endpoint = f"http://{self.host}:{port}{request}"
+
+        return await self.httpx_async_client.async_get(
+            endpoint, self.timeout, self.read_timeout, cache_id=cache_id
         )
 
-        client = self.async_client_getter()
-        try:
-            res = await client.get(endpoint, timeout=self.timeout)
-            res.raise_for_status()
-        finally:
-            # Close the default AsyncClient but keep custom clients open
-            if self.is_default_async_client():
-                await client.aclose()
-
-        return res
-
-    @async_handle_receiver_exceptions
     async def async_post(
         self,
         request: str,
+        *,
         content: Optional[bytes] = None,
         data: Optional[Dict] = None,
         port: Optional[int] = None,
+        cache_id: Hashable = None,
     ) -> httpx.Response:
         """Call POST endpoint of Denon AVR receiver asynchronously."""
         # Use default port of the receiver if no different port is specified
         port = port if port is not None else self.port
 
-        endpoint = "http://{host}:{port}{request}".format(
-            host=self.host, port=port, request=request
+        endpoint = f"http://{self.host}:{port}{request}"
+
+        return await self.httpx_async_client.async_post(
+            endpoint,
+            self.timeout,
+            self.read_timeout,
+            content=content,
+            data=data,
+            cache_id=cache_id,
         )
 
-        client = self.async_client_getter()
-        try:
-            res = await client.post(
-                endpoint, content=content, data=data, timeout=self.timeout
-            )
-            res.raise_for_status()
-        finally:
-            # Close the default AsyncClient but keep custom clients open
-            if self.is_default_async_client():
-                await client.aclose()
-
-        return res
-
-    @async_handle_receiver_exceptions
     async def async_get_command(self, request: str) -> str:
         """Send HTTP GET command to Denon AVR receiver asynchronously."""
         # HTTP GET to endpoint
@@ -159,38 +232,46 @@ class DenonAVRApi:
         # Return text
         return res.text
 
-    @set_cache_id
-    @cache_clear_on_exception
-    @lru_cache(maxsize=32)
-    @async_handle_receiver_exceptions
     async def async_get_xml(
-        self, request: str, cache_id: Hashable = None
+        self, request: str, *, cache_id: Hashable = None
     ) -> ET.Element:
         """Return XML data from HTTP GET endpoint asynchronously."""
         # HTTP GET to endpoint
-        res = await self.async_get(request)
+        res = await self.async_get(request, cache_id=cache_id)
         # create ElementTree
-        xml_root = fromstring(res.text)
+        try:
+            xml_root = fromstring(res.text)
+        except (
+            ET.ParseError,
+            DefusedXmlException,
+            ParseError,
+            UnicodeDecodeError,
+        ) as err:
+            raise AvrInvalidResponseError(f"XMLParseError: {err}", request) from err
         # Check validity of XML
         self.check_xml_validity(request, xml_root)
         # Return ElementTree element
         return xml_root
 
-    @set_cache_id
-    @cache_clear_on_exception
-    @lru_cache(maxsize=32)
-    @async_handle_receiver_exceptions
     async def async_post_appcommand(
-        self, request: str, cmds: Tuple[AppCommandCmd], cache_id: Hashable = None
+        self, request: str, cmds: Tuple[AppCommandCmd], *, cache_id: Hashable = None
     ) -> ET.Element:
         """Return XML from Appcommand(0300) endpoint asynchronously."""
         # Prepare XML body for POST call
         content = self.prepare_appcommand_body(cmds)
         _LOGGER.debug("Content for %s endpoint: %s", request, content)
         # HTTP POST to endpoint
-        res = await self.async_post(request, content=content)
+        res = await self.async_post(request, content=content, cache_id=cache_id)
         # create ElementTree
-        xml_root = fromstring(res.text)
+        try:
+            xml_root = fromstring(res.text)
+        except (
+            ET.ParseError,
+            DefusedXmlException,
+            ParseError,
+            UnicodeDecodeError,
+        ) as err:
+            raise AvrInvalidResponseError(f"XMLParseError: {err}", request) from err
         # Check validity of XML
         self.check_xml_validity(request, xml_root)
         # Add query tags to result
@@ -201,7 +282,7 @@ class DenonAVRApi:
     def add_appcommand_update_tag(self, tag: AppCommandCmd) -> None:
         """Add appcommand tag for full update."""
         if tag.cmd_id != "1":
-            raise ValueError("cmd_id is {} but must be 1".format(tag.cmd_id))
+            raise ValueError(f"cmd_id is {tag.cmd_id} but must be 1")
 
         # Remove response pattern from tag because it is not relevant for query
         tag = attr.evolve(tag, response_pattern=tuple())
@@ -213,7 +294,7 @@ class DenonAVRApi:
     def add_appcommand0300_update_tag(self, tag: AppCommandCmd) -> None:
         """Add appcommand0300 tag for full update."""
         if tag.cmd_id != "3":
-            raise ValueError("cmd_id is {} but must be 3".format(tag.cmd_id))
+            raise ValueError(f"cmd_id is {tag.cmd_id} but must be 3")
 
         # Remove response pattern from tag because it is not relevant for query
         tag = attr.evolve(tag, response_pattern=tuple())
@@ -247,16 +328,20 @@ class DenonAVRApi:
         """
         if len(cmd_list) != len(xml_root):
             raise AvrIncompleteResponseError(
-                "Invalid length of response XML. Query has {} elements, "
-                "response {}".format(len(cmd_list), len(xml_root)),
+                (
+                    "Invalid length of response XML. Query has"
+                    f" {len(cmd_list)} elements, response {len(xml_root)}"
+                ),
                 request,
             )
 
         for i, child in enumerate(xml_root):
             if child.tag not in ["cmd", "error"]:
                 raise AvrInvalidResponseError(
-                    'Returned document contains a tag other than "cmd" and '
-                    '"error": {}'.format(child.tag),
+                    (
+                        'Returned document contains a tag other than "cmd" and'
+                        f' "error": {child.tag}'
+                    ),
                     request,
                 )
             # Find corresponding attributes from request XML if set and add
@@ -332,10 +417,6 @@ class DenonAVRApi:
 
         return body_bytes
 
-    def is_default_async_client(self) -> bool:
-        """Check if default httpx.AsyncCLient getter is used."""
-        return self.async_client_getter is get_default_async_client
-
 
 class DenonAVRTelnetProtocol(asyncio.Protocol):
     """Protocol for the Denon AVR Telnet interface."""
@@ -352,12 +433,17 @@ class DenonAVRTelnetProtocol(asyncio.Protocol):
     @property
     def connected(self) -> bool:
         """Return True if transport is connected."""
-        return self.transport is not None
+        if self.transport is None:
+            return False
+        return not self.transport.is_closing()
 
     def write(self, data: str) -> None:
         """Write data to the transport."""
-        if self.transport is not None:
-            self.transport.write(data.encode("utf-8"))
+        if self.transport is None:
+            return
+        if self.transport.is_closing():
+            return
+        self.transport.write(data.encode("utf-8"))
 
     def close(self) -> None:
         """Close the connection."""
@@ -378,9 +464,9 @@ class DenonAVRTelnetProtocol(asyncio.Protocol):
 
     def connection_lost(self, exc: Optional[Exception]) -> None:
         """Handle connection lost."""
-        self.transport = None
         self._on_connection_lost()
-        return super().connection_lost(exc)
+        super().connection_lost(exc)
+        self.transport = None
 
 
 @attr.s(auto_attribs=True, hash=False, on_setattr=DENON_ATTR_SETATTR)
@@ -390,25 +476,42 @@ class DenonAVRTelnetApi:
     host: str = attr.ib(converter=str, default="localhost")
     timeout: float = attr.ib(converter=float, default=2.0)
     _connection_enabled: bool = attr.ib(default=False)
-    _healthy: Optional[bool] = attr.ib(
-        converter=attr.converters.optional(bool), default=None
-    )
     _last_message_time: float = attr.ib(default=-1.0)
     _connect_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
     _reconnect_task: asyncio.Task = attr.ib(default=None)
     _monitor_handle: asyncio.TimerHandle = attr.ib(default=None)
     _protocol: DenonAVRTelnetProtocol = attr.ib(default=None)
-    _callbacks: Dict[str, Callable] = attr.ib(
+    _telnet_event_map: Dict[str, List] = attr.ib(
+        default=attr.Factory(telnet_event_map_factory)
+    )
+    _callback_tasks: Set[asyncio.Task] = attr.ib(attr.Factory(set))
+    _send_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
+    _send_confirmation_timeout: float = attr.ib(converter=float, default=2.0)
+    _send_confirmation_event: asyncio.Event = attr.ib(
+        default=attr.Factory(asyncio.Event)
+    )
+    _send_confirmation_command: str = attr.ib(converter=str, default="")
+    _send_tasks: Set[asyncio.Task] = attr.ib(attr.Factory(set))
+    _callbacks: Dict[str, List[Coroutine]] = attr.ib(
         validator=attr.validators.instance_of(dict),
         default=attr.Factory(dict),
         init=False,
     )
+    _raw_callbacks: List[Coroutine] = attr.ib(
+        validator=attr.validators.instance_of(list),
+        default=attr.Factory(list),
+        init=False,
+    )
+
+    def __attrs_post_init__(self) -> None:
+        """Initialize special attributes."""
+        self._register_raw_callback(self._async_send_confirmation_callback)
 
     async def async_connect(self) -> None:
         """Connect to the receiver asynchronously."""
         _LOGGER.debug("%s: telnet connecting", self.host)
         async with self._connect_lock:
-            if self.connected is True:
+            if self.connected:
                 return
             await self._async_establish_connection()
 
@@ -428,27 +531,89 @@ class DenonAVRTelnetApi:
                 )
         except asyncio.TimeoutError as err:
             _LOGGER.debug("%s: Timeout exception on telnet connect", self.host)
-            raise AvrTimoutError(
-                "TimeoutException: {}".format(err), "telnet connect"
-            ) from err
+            raise AvrTimoutError(f"TimeoutException: {err}", "telnet connect") from err
         except ConnectionRefusedError as err:
             _LOGGER.debug(
-                "%s: Connection refused on telnet connect", self.host, exc_info=True
+                "%s: Connection refused on telnet connect: %s", self.host, err
             )
             raise AvrNetworkError(
-                "ConnectionRefusedError: {}".format(err), "telnet connect"
+                f"ConnectionRefusedError: {err}", "telnet connect"
             ) from err
         except (OSError, IOError) as err:
             _LOGGER.debug(
-                "%s: Connection failed on telnet reconnect", self.host, exc_info=True
+                "%s: Connection failed on telnet reconnect: %s", self.host, err
             )
-            raise AvrNetworkError("OSError: {}".format(err), "telnet connect") from err
-        _LOGGER.debug("%s: telnet connection complete", self.host)
+            raise AvrNetworkError(f"OSError: {err}", "telnet connect") from err
         self._protocol = cast(DenonAVRTelnetProtocol, transport_protocol[1])
+        _LOGGER.debug("%s: telnet connection established", self.host)
         self._connection_enabled = True
         self._last_message_time = time.monotonic()
         self._schedule_monitor()
-        self._protocol.write("PW?\r")
+        # Trigger update of all attributes
+        await self.async_send_commands(
+            "ZM?",
+            "SI?",
+            "MV?",
+            "MU?",
+            "Z2?",
+            "Z2MU?",
+            "Z3?",
+            "Z3MU?",
+            "PSTONE CTRL ?",
+            "PSBAS ?",
+            "PSTRE ?",
+            "PSDYNEQ ?",
+            "PSLFC ?",
+            "PSCNTAMT ?",
+            "PSMULTEQ: ?",
+            "PSREFLEV ?",
+            "PSDYNVOL ?",
+            "MS?",
+            "MNMEN?",
+            "DIM ?",
+            "PSDELAY?",
+            "ECO?",
+            "VSMONI ?",
+            "PSDIRAC ?",
+            "CV?",
+            "PSNEURAL ?",
+            "PSIMAX ?",
+            "PSIMAXAUD ?",
+            "PSIMAXHPF ?",
+            "PSIMAXLPF ?",
+            "PSIMAXSWM ?",
+            "PSIMAXSWO ?",
+            "PSSWR ?",
+            "PSSWL ?",
+            "SSTTR ?",
+            "MSQUICK ?",
+            "STBY?",
+            "SLP?",
+            "VSAUDIO ?",
+            "PSCES ?",
+            "VSVPM ?",
+            "PSLFE ?",
+            "PSLOM ?",
+            "PSBSC ?",
+            "PSDEH ?",
+            "PSCINEMA EQ. ?",
+            "PSAUROPR ?",
+            "PSAUROST ?",
+            "PSAUROMODE ?",
+            "PSRSZ ?",
+            "TR?",
+            "SPPR ?",
+            "BTTX ?",
+            "PSDIC ?",
+            "PSSPV ?",
+            "PSSP: ?",
+            "PSDRC ?",
+            "PSDEL ?",
+            "PSRSTR ?",
+            "PSGEQ ?",
+            "PSHEQ ?",
+            skip_confirmation=True,
+        )
 
     def _schedule_monitor(self) -> None:
         """Start the monitor task."""
@@ -468,8 +633,6 @@ class DenonAVRTelnetApi:
             _LOGGER.info(
                 "%s: Keep alive failed, disconnecting and reconnecting", self.host
             )
-            if self._protocol:
-                self._protocol.close()
             self._handle_disconnected()
             return
 
@@ -481,16 +644,20 @@ class DenonAVRTelnetApi:
 
     def _handle_disconnected(self) -> None:
         """Handle disconnected."""
-        _LOGGER.debug("%s: disconnected", self.host)
-        self._protocol = None
+        _LOGGER.debug("%s: handle disconnected", self.host)
+        if self._protocol is not None:
+            self._protocol.close()
+            self._protocol = None
         self._stop_monitor()
         if not self._connection_enabled:
             return
-        self._reconnect_task = asyncio.create_task(self._async_reconnect())
+        if self._reconnect_task is None:
+            self._reconnect_task = asyncio.create_task(self._async_reconnect())
 
     async def async_disconnect(self) -> None:
         """Close the connection to the receiver asynchronously."""
         async with self._connect_lock:
+            _LOGGER.debug("%s: telnet disconnecting", self.host)
             self._connection_enabled = False
             self._stop_monitor()
             reconnect_task = self._reconnect_task
@@ -506,44 +673,56 @@ class DenonAVRTelnetApi:
                     await reconnect_task
                 except asyncio.CancelledError:
                     pass
+            _LOGGER.debug("%s: telnet disconnected", self.host)
 
     async def _async_reconnect(self) -> None:
         """Reconnect to the receiver asynchronously."""
         backoff = 0.5
 
-        while self._connection_enabled is True and not self.healthy:
+        while self._connection_enabled and not self.healthy:
             async with self._connect_lock:
+                _LOGGER.debug("%s: Telnet reconnecting", self.host)
                 try:
                     await self._async_establish_connection()
                 except AvrTimoutError:
                     _LOGGER.debug(
                         "%s: Timeout exception on telnet reconnect", self.host
                     )
-                except AvrNetworkError as ex:
-                    _LOGGER.debug("%s: %s", self.host, ex, exc_info=True)
-                except Exception:  # pylint: disable=broad-except
+                except AvrNetworkError as err:
+                    _LOGGER.debug("%s: %s", self.host, err)
+                except AvrProcessingError as err:
+                    _LOGGER.debug(
+                        "%s: Failed updating state on telnet reconnect: %s",
+                        self.host,
+                        err,
+                    )
+                except Exception as err:  # pylint: disable=broad-except
                     _LOGGER.error(
                         "%s: Unexpected exception on telnet reconnect",
                         self.host,
-                        exc_info=True,
+                        exc_info=err,
                     )
                 else:
                     _LOGGER.info("%s: Telnet reconnected", self.host)
-                    return
+                    break
 
             await asyncio.sleep(backoff)
             backoff = min(30.0, backoff * 2)
 
+        self._reconnect_task = None
+
     def register_callback(
         self, event: str, callback: Callable[[str, str, str], Awaitable[None]]
     ) -> None:
         """Register a callback handler for an event type."""
         # Validate the passed in type
-        if event != "ALL" and event not in TELNET_EVENTS:
-            raise ValueError("{} is not a valid callback type.".format(event))
+        if event != ALL_TELNET_EVENTS and event not in TELNET_EVENTS:
+            raise ValueError(f"{event} is not a valid callback type.")
 
         if event not in self._callbacks.keys():
             self._callbacks[event] = []
+        elif callback in self._callbacks[event]:
+            return
         self._callbacks[event].append(callback)
 
     def unregister_callback(
@@ -554,18 +733,31 @@ class DenonAVRTelnetApi:
             return
         self._callbacks[event].remove(callback)
 
+    def _register_raw_callback(
+        self, callback: Callable[[str], Awaitable[None]]
+    ) -> None:
+        """Register a callback handler for raw telnet messages."""
+        if callback in self._raw_callbacks:
+            return
+        self._raw_callbacks.append(callback)
+
+    def _unregister_raw_callback(
+        self, callback: Callable[[str], Awaitable[None]]
+    ) -> None:
+        """Unregister a callback handler for raw telnet messages."""
+        self._raw_callbacks.remove(callback)
+
     def _process_event(self, message: str) -> None:
         """Process a realtime event."""
         _LOGGER.debug("Incoming Telnet message: %s", message)
         self._last_message_time = time.monotonic()
         if len(message) < 3:
             return
-        zone = MAIN_ZONE
 
         # Event is 2 characters
-        event = message[0:2]
+        event = self._get_event(message)
         # Parameter is the remaining characters
-        parameter = message[2:]
+        parameter = message[len(event) :]
 
         if event == "MV":
             # This seems undocumented by Denon and appears to basically be a
@@ -574,29 +766,49 @@ class DenonAVRTelnetApi:
             if parameter[0:3] == "MAX":
                 return
 
-        if event in ("Z2", "Z3"):
+        # Determine zone
+        zone = MAIN_ZONE
+        if event in ALL_ZONE_TELNET_EVENTS:
+            zone = ALL_ZONES
+        elif event in {"Z2", "Z3"}:
             if event == "Z2":
                 zone = ZONE2
             else:
                 zone = ZONE3
 
-            if parameter in ("ON", "OFF"):
-                event = "PW"
-            elif parameter in TELNET_SOURCES:
+            if parameter in TELNET_SOURCES:
                 event = "SI"
             elif parameter.isdigit():
                 event = "MV"
-            elif parameter[0:2] in TELNET_EVENTS:
-                event = parameter[0:2]
-                parameter = parameter[2:]
+            elif self._get_event(parameter):
+                event = self._get_event(parameter)
+                parameter = parameter[len(event) :]
 
         if event not in TELNET_EVENTS:
             return
 
-        asyncio.create_task(self._async_run_callbacks(event, zone, parameter))
+        task = asyncio.create_task(
+            self._async_run_callbacks(message, event, zone, parameter)
+        )
+        self._callback_tasks.add(task)
+        task.add_done_callback(self._callback_tasks.discard)
+
+    async def _async_run_callbacks(
+        self, message: str, event: str, zone: str, parameter: str
+    ) -> None:
+        """Handle triggering the registered callbacks."""
+        for callback in self._raw_callbacks:
+            try:
+                await callback(message)
+            except Exception as err:  # pylint: disable=broad-except
+                # We don't want a single bad callback to trip up the
+                # whole system and prevent further execution
+                _LOGGER.error(
+                    "%s: Raw callback caused an unhandled exception: %s",
+                    self.host,
+                    err,
+                )
 
-    async def _async_run_callbacks(self, event: str, zone: str, parameter: str) -> None:
-        """Handle triggering the registered callbacks for the event."""
         if event in self._callbacks.keys():
             for callback in self._callbacks[event]:
                 try:
@@ -605,28 +817,83 @@ class DenonAVRTelnetApi:
                     # We don't want a single bad callback to trip up the
                     # whole system and prevent further execution
                     _LOGGER.error(
-                        "%s: Event callback caused an unhandled exception %s",
+                        "%s: Event callback caused an unhandled exception: %s",
                         self.host,
                         err,
                     )
 
-        if "ALL" in self._callbacks.keys():
-            for callback in self._callbacks["ALL"]:
+        if ALL_TELNET_EVENTS in self._callbacks.keys():
+            for callback in self._callbacks[ALL_TELNET_EVENTS]:
                 try:
                     await callback(zone, event, parameter)
                 except Exception as err:  # pylint: disable=broad-except
                     # We don't want a single bad callback to trip up the
                     # whole system and prevent further execution
                     _LOGGER.error(
-                        "%s: Event callback caused an unhandled exception %s",
+                        "%s: Event callback caused an unhandled exception: %s",
                         self.host,
                         err,
                     )
 
-    def send_commands(self, *commands: str) -> None:
+    def _get_event(self, message: str) -> str:
+        """Get event of a telnet message."""
+        events = self._telnet_event_map.get(message[0:2], [""])
+        for event in events:
+            if message.startswith(event):
+                return event
+        return ""
+
+    async def _async_send_confirmation_callback(self, message: str) -> None:
+        """Confirm that the telnet command has been executed."""
+        if len(message) < 3:
+            return
+        command = self._send_confirmation_command
+        if self._get_event(message) == self._get_event(self._send_confirmation_command):
+            self._send_confirmation_command = ""
+            self._send_confirmation_event.set()
+            _LOGGER.debug("Command %s confirmed", command)
+
+    async def _async_send_command(
+        self, command: str, skip_confirmation: bool = False
+    ) -> None:
+        """Send one telnet command to the receiver."""
+        async with self._send_lock:
+            if not skip_confirmation:
+                self._send_confirmation_command = command
+                self._send_confirmation_event.clear()
+            if not self.connected or not self.healthy:
+                raise AvrProcessingError(
+                    f"Error sending command {command}. Telnet connected: "
+                    f"{self.connected}, Connection healthy: {self.healthy}"
+                )
+            self._protocol.write(f"{command}\r")
+            if not skip_confirmation:
+                try:
+                    await asyncio.wait_for(
+                        self._send_confirmation_event.wait(),
+                        self._send_confirmation_timeout,
+                    )
+                except asyncio.TimeoutError:
+                    _LOGGER.info(
+                        "Timeout waiting for confirmation of command: %s", command
+                    )
+                finally:
+                    self._send_confirmation_command = ""
+
+    async def async_send_commands(
+        self, *commands: str, skip_confirmation: bool = False
+    ) -> None:
         """Send telnet commands to the receiver."""
         for command in commands:
-            self._protocol.write("{}\r".format(command))
+            await self._async_send_command(command, skip_confirmation=skip_confirmation)
+
+    def send_commands(self, *commands: str, skip_confirmation: bool = False) -> None:
+        """Send telnet commands to the receiver."""
+        task = asyncio.create_task(
+            self.async_send_commands(*commands, skip_confirmation=skip_confirmation)
+        )
+        self._send_tasks.add(task)
+        task.add_done_callback(self._send_tasks.discard)
 
     ##############
     # Properties #
@@ -637,6 +904,6 @@ class DenonAVRTelnetApi:
         return self._connection_enabled
 
     @property
-    def healthy(self) -> Optional[bool]:
+    def healthy(self) -> bool:
         """Return True if telnet connection is healthy."""
         return self._protocol is not None and self._protocol.connected
