@@ -229,6 +229,10 @@ class DenonAVRSoundMode(DenonAVRFoundation):
         if self._device.zone != zone:
             return
 
+        # QUICK is not a sound mode
+        if parameter.startswith("QUICK"):
+            return
+
         self._sound_mode_raw = parameter
 
     async def _async_neural_x_callback(
@@ -1209,17 +1213,6 @@ class DenonAVRSoundMode(DenonAVRFoundation):
                 )
             )
 
-    async def async_effect_speaker_selection_toggle(self) -> None:
-        """
-        Toggle Effect Speaker Selection.
-
-        Only available if using Telnet.
-        """
-        if self._effect_speaker_selection == "Floor":
-            await self.async_effect_speaker_selection("Height + Floor")
-        else:
-            await self.async_effect_speaker_selection("Floor")
-
     async def async_drc(self, mode: DRCs) -> None:
         """Set DRC mode."""
         if mode not in self._drcs:
